#ifndef slic3r_SupportMaterial_hpp_
#define slic3r_SupportMaterial_hpp_

namespace Slic3r {

// how much we extend support around the actual contact area
constexpr coordf_t SUPPORT_MATERIAL_MARGIN = 1.5;

}

#endif
